#include "glwidget.h"

#include <QOpenGLTexture>
#include <QOpenGLPixelTransferOptions>
#include <QOpenGLBuffer>
#include <QOpenGLShaderProgram>
#include <QPainter>

struct VertexData {
    QVector3D coord;
    QVector2D texCoord;
};

GLWidget::GLWidget(QWidget *parent) :
    QOpenGLWidget(parent),
    m_program(0),
    m_IBO(0),
    m_vertexDataVBO(0),
    m_texture(0),
    m_isRotating(true) {
    startTimer( 10 );
    QFont f = font();
    f.setPointSize( 100 );
    setFont( f );
}

GLWidget::~GLWidget() {
    makeCurrent();
    delete m_program;
    delete m_IBO;
    delete m_vertexDataVBO;
    delete m_texture;
    doneCurrent();
}

void GLWidget::initializeGL() {
    // Creiamo l'oggetto QOpenGLShaderProgram
    m_program = new QOpenGLShaderProgram(this);

    // Compiliamo lo shader dei vertici
    if (!m_program->addShaderFromSourceFile(QOpenGLShader::Vertex, ":/vertex.glsl"))
        close();

    // Compiliamo il fragment shader
    if (!m_program->addShaderFromSourceFile(QOpenGLShader::Fragment, ":/fragment.glsl"))
        close();

    // Linkiamo gli shaders
    if (!m_program->link())
        close();

    // **** Crea i Vertex Buffer Object ****

    // *** VBO dat vertici ***
    // vettore contenete i dati associati ai vertici
    float l = 1.0 / 2.0;
    VertexData vertexData[] = {
        // faccia 0 - 1
        {QVector3D(-l, -l,  l), QVector2D(0.0f, 0.0f)},
        {QVector3D( l, -l,  l), QVector2D(1.0f, 0.0f)},
        {QVector3D(-l,  l,  l), QVector2D(0.0f, 1.0f)},
        {QVector3D( l,  l,  l), QVector2D(1.0f, 1.0f)},

        // faccia 1 - 2
        {QVector3D( l, -l,  l), QVector2D(0.0f, 0.0f)},
        {QVector3D( l, -l, -l), QVector2D(1.0f, 0.0f)},
        {QVector3D( l,  l,  l), QVector2D(0.0f, 1.0f)},
        {QVector3D( l,  l, -l), QVector2D(1.0f, 1.0f)},

        // faccia 2 - 6
        {QVector3D( l, -l, -l), QVector2D(0.0f, 0.0f)},
        {QVector3D(-l, -l, -l), QVector2D(1.0f, 0.0f)},
        {QVector3D( l,  l, -l), QVector2D(0.0f, 1.0f)},
        {QVector3D(-l,  l, -l), QVector2D(1.0f, 1.0f)},

        // faccia 3 - 5
        {QVector3D(-l, -l, -l), QVector2D(0.0f, 0.0f)},
        {QVector3D(-l, -l,  l), QVector2D(1.0f, 0.0f)},
        {QVector3D(-l,  l, -l), QVector2D(0.0f, 1.0f)},
        {QVector3D(-l,  l,  l), QVector2D(1.0f, 1.0f)},

        // faccia 4 - 3
        {QVector3D(-l, -l, -l), QVector2D(0.0f, 0.0f)},
        {QVector3D( l, -l, -l), QVector2D(1.0f, 0.0f)},
        {QVector3D(-l, -l,  l), QVector2D(0.0f, 1.0f)},
        {QVector3D( l, -l,  l), QVector2D(1.0f, 1.0f)},

        // faccia 5 - 4
        {QVector3D(-l,  l,  l), QVector2D(0.0f, 0.0f)},
        {QVector3D( l,  l,  l), QVector2D(1.0f, 0.0f)},
        {QVector3D(-l,  l, -l), QVector2D(0.0f, 1.0f)},
        {QVector3D( l,  l, -l), QVector2D(1.0f, 1.0f)}
    };
    m_vertexDataVBO = new QOpenGLBuffer(QOpenGLBuffer::VertexBuffer);
    m_vertexDataVBO->create();
    m_vertexDataVBO->setUsagePattern( QOpenGLBuffer::StaticDraw );
    m_vertexDataVBO->bind();
    m_vertexDataVBO->allocate( vertexData, sizeof(vertexData) );
    m_vertexDataVBO->release();

    // *** texture ***
    QImage img = drawImage("8");
    m_texture = new QOpenGLTexture( img );

    // Parametri per la gestione della texture negli shader
    // Ripete l'immagine se necessario
    m_texture->setWrapMode(QOpenGLTexture::Repeat);
    // quando l'area è più grande della texture, interpola linearmente l'originale
    m_texture->setMagnificationFilter(QOpenGLTexture::Linear);
    // quando l'area è più piccola della texture, interpola linearmente l'originale
    m_texture->setMinificationFilter(QOpenGLTexture::Linear);


    // *** IBO ***
    GLushort indices[] = { 0,  1,  2,  1,  2,  3,  // Faccia 0
                           4,  5,  6,  5,  6,  7,  // Faccia 1
                           8,  9, 10,  9, 10,  11, // Faccia 2
                           12, 13, 14, 13, 14, 15, // Faccia 3
                           16, 17, 18, 17, 18, 19, // Faccia 4
                           20, 21, 22, 21, 22, 23  // Faccia 5
                         };
    m_IBO = new QOpenGLBuffer(QOpenGLBuffer::IndexBuffer);
    m_IBO->create();
    m_IBO->setUsagePattern( QOpenGLBuffer::StaticDraw );
    m_IBO->bind();
    m_IBO->allocate( indices, sizeof(indices) );
    m_IBO->release();

    // Attiviamo alcuni stati
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_BLEND);
    glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    // Localizzaziamo gli attributi negli shader
    m_vertexPos = m_program->attributeLocation("vertex");
    m_texCoordPos = m_program->attributeLocation( "texCoord" );
    m_matrixPos = m_program->uniformLocation("matrix");
    m_texturePos = m_program->uniformLocation("texture");

    // impostiamo la prospettiva
    m_matrix.perspective(60, 4.0/3.0, 0.1, 100.0);
    m_matrix.translate(0, 0, -2);
}

void GLWidget::resizeGL(int w, int h) {
    glViewport(0, 0, w, h);
}

void GLWidget::paintGL() {
    // **** Disegni opengl ****

    // Rendiamo corrente il contesto - in questo caso non è
    // necessario, ma è una buona abitudine
    makeCurrent();

    // Puliamo la la vista
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);

    // Linkiamo gli shader
    m_program->bind();

    // Carichiamo la matrice modello/vista
    m_program->setUniformValue(m_matrixPos, m_matrix);

    // Attiviamo il VBO dei vertici
    m_vertexDataVBO->bind();

    m_program->enableAttributeArray( m_vertexPos );
    GLsizei stride = sizeof(VertexData);        // distanza tra l'inizio di un vertice e l'inizio dell'altro
    m_program->setAttributeBuffer( m_vertexPos, GL_FLOAT, 0, 3, stride);

    m_program->enableAttributeArray( m_texCoordPos );
    m_program->setAttributeBuffer( m_texCoordPos, GL_FLOAT, sizeof(QVector3D), 2, stride);

    m_vertexDataVBO->release();

    // *** texture ***
    // attiviamo l'unita' 0
    glActiveTexture(GL_TEXTURE0);
    // attiviamo la texture
    m_texture->bind();
    // impostiamo il valore nell'unita' 0
    m_program->setUniformValue( m_texturePos, 0);

    m_IBO->bind();

    // Disegna i triangoli
    // Disegna i triangoli
    glDrawElements( GL_TRIANGLES,       // modalita
                    36,                  // numero indici
                    GL_UNSIGNED_SHORT,  // tipologia degli indici
                    0                   // offset tra gli indici
                    );

    // Rilasciamo gli shader
    m_program->release();
}

void GLWidget::timerEvent(QTimerEvent *event) {
    Q_UNUSED(event);

    if( m_isRotating ){
        // Ruotiamo di 1 grado attorno all'asse y nel sistema di riferimento dell'osservatore
        m_matrix.rotate(1.0, 0.0, 1.0, 0.0 );
        m_matrix.rotate(0.5, 1.0, 0.0, 0.0 );

        // Aggiorniamo il disegno
        update();
    }
}

void GLWidget::mousePressEvent(QMouseEvent *event){
    Q_UNUSED(event);

    // Invertiamo il valore di m_isRotating
    m_isRotating = !m_isRotating;
}

QImage GLWidget::drawImage( const QString & txt ){
    int border = 10;
    int side = qMax( fontMetrics().width(txt) + border , fontMetrics().height()  + border );
    QPixmap pixmap( side, side );
    pixmap.fill(Qt::white);
    QPainter painter;
    painter.begin(&pixmap);
    painter.setRenderHints(QPainter::HighQualityAntialiasing | QPainter::TextAntialiasing);
    painter.setFont(font());
    painter.setPen( Qt::red );
    painter.drawText(QRect(0,0,side,side), Qt::AlignCenter, txt );
    painter.end();

    return pixmap.toImage().convertToFormat( QImage::Format_RGBA8888 );
}
